from typing import Optional, Annotated
from datetime import timedelta
import json

from pydantic import field_validator
from pydantic_geojson import PointModel, PolygonModel
from sqlmodel import Field, SQLModel, Column, Relationship
from sqlmodel import JSON, ARRAY, String, Text
from pydantic import UUID4, HttpUrl
from uuid import uuid4
from geoalchemy2 import Geometry, WKBElement
from geoalchemy2.shape import to_shape
from shapely import to_geojson, from_geojson
from shapely.wkt import loads, dumps
from datetime import datetime
from dateutil.relativedelta import relativedelta


def get_embargo_end(value=None):
    if value is None:
        value = datetime.now()
    return value + relativedelta(years=2)


class PersonRole(SQLModel, table=True):
    __tablename__ = 'person_roles'

    # columns
    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(nullable=False, unique=True)
    description: str | None = None


class NMPersonEntries(SQLModel, table=True):
    __tablename__ = 'nm_persons_entries'

    person_id: int = Field(foreign_key='persons.id', primary_key=True)
    entry_id: int = Field(foreign_key='entries.id', primary_key=True)
    relationship_type_id: int = Field(default=13, foreign_key='person_roles.id')
    order: int = Field(default=1)

class PersonBase(SQLModel):
    is_organisation: bool = False
    first_name: str | None = None
    last_name: str | None = None
    organisation_name: str | None = None
    organisation_abbrev: str | None = None
    affiliation: str | None = None
    attribution: str | None = None

class PersonTable(PersonBase, table=True):
    __tablename__ = 'persons'
    id: int | None = Field(default=None, primary_key=True)
    uuid: UUID4 = Field(default_factory=uuid4)
    entries: list['EntryTable'] = Relationship(back_populates='author')

class Author(PersonBase):
    id: int
    uuid: UUID4

class CoAuthor(PersonBase):
    id: int
    uuid: UUID4

class AuthorCreate(PersonBase):
    pass

class LicenseBase(SQLModel):
    short_title: str = Field(nullable=False, unique=True)
    title: str = Field(nullable=False)
    summary: str
    full_text: str | None = None
    link: str | None = None
    by_attribution: bool = False
    share_alike: bool = False
    commercial_use: bool = False

class LicenseTable(LicenseBase, table=True):
    __tablename__ = 'licenses'
    id: int | None = Field(default=None, primary_key=True) 
    entries: list['EntryTable'] = Relationship(back_populates='license')

class LicenseCreate(LicenseBase):
    pass

class License(LicenseBase):
    id: int

class ThesaurusBase(SQLModel):
    uuid: UUID4 = Field(default_factory=uuid4)
    name: str = Field(nullable=False, unique=True)
    title: str = Field(nullable=False)
    organisation: str = Field(nullable=False)
    description: str | None = None
    url: HttpUrl = Field(sa_column=Column(Text, nullable=False))

class ThesaurusTable(ThesaurusBase, table=True):
    __tablename__ = 'thesaurus'
    id: int | None = Field(default=None, primary_key=True)

class Thesaurus(ThesaurusBase):
    id: int

class NMKeywordsEntries(SQLModel, table=True):
    __tablename__ = 'nm_keywords_entries'
    keyword_id: int = Field(foreign_key='keywords.id', primary_key=True)
    entry_id: int = Field(foreign_key='entries.id', primary_key=True)


class KeywordBase(SQLModel):
    id: int | None = None
    uuid: UUID4 | None = None
    parent_id: int | None = Field(foreign_key='keywords.id')
    value: str = Field(nullable=False)
    full_path: str | None = None

class KeywordTable(KeywordBase, table=True):
    __tablename__ = 'keywords'
    id: int = Field(primary_key=True)
    uuid: UUID4 = Field(default_factory=uuid4)
    thesaurus_id: int = Field(foreign_key='thesaurus.id')

    # relationships
    thesaurus: ThesaurusTable = Relationship()
    entries: list['EntryTable'] = Relationship(back_populates='keywords', link_model=NMKeywordsEntries)

class Keyword(KeywordBase):
    thesaurus: ThesaurusBase


class UnitBase(SQLModel):
    name: str = Field(nullable=False, unique=True)
    symbol: str = Field(nullable=False)
    si: str | None = None

class UnitTable(UnitBase, table=True):
    __tablename__ = 'units'
    id: int | None = Field(default=None, primary_key=True)
    variables: list['VariableTable'] = Relationship(back_populates='unit')

class Unit(UnitBase):
    pass

class VariableBase(SQLModel):
    name: str = Field(nullable=False, unique=True)
    symbol: str = Field(nullable=False)
    column_names: list[str] = Field(sa_column=Column(ARRAY(String)))

class VariableTable(VariableBase, table=True):
    __tablename__ = 'variables'

    # columns
    id: int | None = Field(default=None, primary_key=True)
    unit_id: int = Field(foreign_key='units.id')
    keyword_id: int | None = Field(foreign_key='keywords.id')

    # relationships
    unit: UnitTable = Relationship(back_populates='variables')
    entries: list['EntryTable'] = Relationship(back_populates='variable')
    keyword: list[KeywordTable] = Relationship()

class Variable(VariableBase):
    id: int
    unit: Unit
    keyword: Keyword | None = None

class DetailBase(SQLModel):
    key: str
    stem: str | None = None
    title: str | None = None
    raw_value: dict = Field(sa_column=Column(JSON), default={})
    description: str | None = None

    @field_validator('raw_value', mode='before')
    def validate_raw_value(cls, v, info):
        if v is None:
            return {}
        elif not isinstance(v, dict):
            return {'__literal__': v}
        else:
            return v

class DetailTable(DetailBase, table=True):
    __tablename__ = 'details'
    id: int | None = Field(default=None, primary_key=True)
    entry_id: int = Field(foreign_key='entries.id')
    thesaurus_id: int | None = Field(foreign_key='thesaurus.id', default=None)

    entry: 'EntryTable' = Relationship(back_populates='details')
    thesaurus: ThesaurusTable = Relationship()


class Detail(DetailBase):
    thesaurus: ThesaurusBase | None = None

    @property
    def value(self):
        # check if the raw_value has a literal value
        if '__literal__' in self.raw_value:
            return self.raw_value['__literal__']
        else:
            return self.raw_value

class DetailCreate(DetailBase):
    thesaurus: int | None = None

class DatasourceTypeBase(SQLModel):
    id: int | None = None
    name: str
    title: str
    description: str | None = None

class DatasourceTypeTable(DatasourceTypeBase, table=True):
    __tablename__ = 'datasource_types'
    id: int = Field(primary_key=True)
    datasources: list['DatasourceTable'] = Relationship(back_populates='type')

class TemporalScaleBase(SQLModel):
    id: int | None = None
    resolution: timedelta
    observation_start: datetime
    observation_end: datetime
    support: float
    dimension_names: list[str] = Field(sa_column=Column(ARRAY(String)))

class TemporalScaleTable(TemporalScaleBase, table=True):
    __tablename__ = 'temporal_scales'
    id: int | None = Field(default=None, primary_key=True)
    datasources: list['DatasourceTable'] = Relationship(back_populates='temporal_scale')

class TemporalScale(TemporalScaleBase):
    @property
    def extent(self):
        return [self.observation_start, self.observation_end]


class SpatialScaleBase(SQLModel):
    id: int | None = None
    resolution: int
    extent: Optional[Annotated[PolygonModel, Field(sa_column=Column(Geometry(geometry_type='POLYGON', srid=4326), nullable=True))]] = None
    support: float
    dimension_names: list[str] = Field(sa_column=Column(ARRAY(String)))

    @field_validator('extent', mode='before')
    def validate_extent(cls, v, info):
        if v is None:
            return None
        elif isinstance(v, str):
            shape = loads(v)
            js = json.loads(to_geojson(shape))
            poly = PolygonModel.model_validate(js)
        elif isinstance(v, (list, tuple)):
            poly = PolygonModel(coordinates=v)
        elif isinstance(v, WKBElement):
            shape = to_shape(v)
            js = json.loads(to_geojson(shape))
            poly = PolygonModel.model_validate(js)
        elif isinstance(v, dict):
            poly = PolygonModel(**v)
        elif isinstance(v, PolygonModel):
            poly = v
        else:
            raise ValueError(f"Invalid geometry type: {v} ({type(v)})")

        return poly

class SpatialScaleTable(SpatialScaleBase, table=True):
    __tablename__ = 'spatial_scales'
    id: int | None = Field(default=None, primary_key=True)
    extent: Optional[str] = Field(sa_column=Column(Geometry(geometry_type='POLYGON', srid=4326), nullable=True))
    datasources: list['DatasourceTable'] = Relationship(back_populates='spatial_scale')

    @field_validator('extent', mode='before')
    def validate_extent(cls, v, info):
        if v is None:
            return None
        elif isinstance(v, str):
            shape = loads(v)
        elif isinstance(v, PolygonModel):
            shape = from_geojson(v.model_dump_json())
        elif isinstance(v, WKBElement):
            shape = to_shape(v)
        else:
            raise ValueError(f"Can't transform extent back to WKBElement: {v} ({type(v)})")
        return  dumps(shape)

class SpatialScale(SpatialScaleBase):
    pass

class DatasourceBase(SQLModel):
    path: str
    encoding: str = 'utf-8'
    variable_names: list[str] = Field(sa_column=Column(ARRAY(String)))
    args: dict | None = Field(sa_column=Column(JSON))

class DatasourceTable(DatasourceBase, table=True):
    __tablename__ = 'datasources'
    
    # columns
    id: int | None = Field(default=None, primary_key=True)
    type_id: int = Field(foreign_key='datasource_types.id')
    temporal_scale_id: int | None = Field(foreign_key='temporal_scales.id')
    spatial_scale_id: int | None = Field(foreign_key='spatial_scales.id')

    # relationships
    type: DatasourceTypeTable = Relationship(back_populates='datasources')
    temporal_scale: TemporalScaleTable = Relationship(back_populates='datasources')
    spatial_scale: SpatialScaleTable = Relationship(back_populates='datasources')
    entry: 'EntryTable' = Relationship(back_populates='datasource')

class DatasourceCreate(DatasourceBase):
    type: int | str
    temporal_scale: TemporalScale | None = None
    spatial_scale: SpatialScale | None = None

class Datasource(DatasourceBase):
    id: int
    type: DatasourceTypeBase
    temporal_scale: TemporalScale | None = None
    spatial_scale: SpatialScale | None = None

class EntryBase(SQLModel):
    title: str = Field(nullable=False, unique=True)
    abstract: str
    external_id: str | None = None
    location: Optional[Annotated[PointModel, Field(sa_column=Column(Geometry(geometry_type='POINT', srid=4326), nullable=True, default=None))]] = None
    version: int  = 1
    latest_version_id: int | None = None
    is_partial: bool = False
    comment: str | None = None
    citation: str | None = None
    embargo: bool = False
    embargo_end: datetime | None = None
    publication: datetime | None = None
    lastUpdate: datetime | None = None
    
    @field_validator('location', mode='before')
    def validate_location(cls, v, info):
        if v is None:
            return None
        elif isinstance(v, (list, tuple)):
            point = PointModel(coordinates=v)
        elif isinstance(v, WKBElement):
            shape = to_shape(v)
            js = json.loads(to_geojson(shape))
            point = PointModel.model_validate(js)
        elif isinstance(v, dict):
            point = PointModel(**v)
        elif isinstance(v, str):
            shape = loads(v)
            js = json.loads(to_geojson(shape))
            point = PointModel.model_validate(js)
        elif isinstance(v, PointModel):
            point = v
        else:
            raise ValueError(f"Invalid location type: {v} ({type(v)})")
        
        return point

class EntryTable(EntryBase, table=True):
    __tablename__ = 'entries'

    # columns
    id: int | None = Field(default=None, primary_key=True)
    uuid: UUID4 = Field(default_factory=uuid4)
    location: Optional[str] = Field(sa_column=Column(Geometry(geometry_type='POINT', srid=4326), nullable=True, default=None))
    author_id: int | None = Field(foreign_key='persons.id')
    license_id: int | None = Field(default=None, foreign_key="licenses.id")
    variable_id: int = Field(foreign_key='variables.id')
    datasource_id: int | None = Field(foreign_key = 'datasources.id')
    
    # add the datetime fields
    embargo_end: datetime = Field(default_factory=get_embargo_end)
    publication: datetime = Field(default_factory=datetime.now)
    lastUpdate: datetime = Field(default_factory=datetime.now)


    # relationships
    license: LicenseTable = Relationship(back_populates='entries')
    author: PersonTable = Relationship()
    coAuthors: list[PersonTable] = Relationship(link_model=NMPersonEntries)
    variable: VariableTable = Relationship(back_populates='entries')
    keywords: list[KeywordTable] = Relationship(back_populates='entries', link_model=NMKeywordsEntries)
    details: list[DetailTable] = Relationship(back_populates='entry')
    datasource: DatasourceTable = Relationship(back_populates='entry')

    @field_validator('location', mode='before')
    def validate_location(cls, v, info):
        if v is None:
            return None
        elif isinstance(v, str):
            shape = loads(v)
        elif isinstance(v, dict):
            shape = from_geojson(PointModel.model_validate(v).model_dump_json())
        elif isinstance(v, PointModel):
            shape = from_geojson(v.model_dump_json())
        elif isinstance(v, WKBElement):
            shape = to_shape(v)
        else:
            raise ValueError(f"Can't transform location back to WKBElement: {v} ({type(v)})")
        return  dumps(shape)


class EntryCreate(EntryBase):
    license: LicenseCreate | int
    variable: int
    author: AuthorCreate | int
    coAuthors: list[AuthorCreate | int] | None = None
    keywords: list[int] | None = None
    details: list[DetailCreate] | None = None
    datasource: Datasource | None = None


class Metadata(EntryBase):
    id: int
    uuid: UUID4
    license: License | None
    variable: Variable
    author: Author
    coAuthors: list[CoAuthor] = []
    keywords: list[Keyword] = []
    details: list[Detail] = []
    datasource: Datasource | None = None
    # associated_groups: List['EntryGroup'] = relationship("EntryGroup", secondary="nm_entrygroups", back_populates='entries')




from typing import Literal, Optional, List
from pathlib import Path
from contextlib import asynccontextmanager
import logging

from fastapi import FastAPI, HTTPException, Request
from fastapi.templating import Jinja2Templates
from fastapi.staticfiles import StaticFiles
from pydantic_geojson import FeatureCollectionModel

from metacatalog_api import core
from metacatalog_api.views import editing_tools
from metacatalog_api import models


logger = logging.getLogger('uvicorn.error')

# before we initialize the app, we check that the database is installed and up to date
@asynccontextmanager
async def lifespan(app: FastAPI):
    # check if the entries table can be found in the database
    with core.connect() as session:
        if not core.db.check_installed(session):
            logger.info("Database not installed, installing...")
            core.db.install(session, populate_defaults=True)
            logger.info("Database installed.")
    
    # after checking the database, we check the version
    with core.connect() as session:
        core.db.check_db_version(session)

    # now we yield the application
    yield

    # here we can app tear down code - i.e. a log message

# define the format literal
FMT = Optional[Literal['html', 'json']]


# build the base app
app = FastAPI(lifespan=lifespan)

# add the templates
templates = Jinja2Templates(directory=Path(__file__).parent / "templates")


@app.get('/')
def index(request: Request):
    return templates.TemplateResponse(request=request, name="index.html", context={})


@app.get('/entries')
@app.get('/entries.{fmt}')
def get_entries(request: Request, fmt: FMT = None, offset: int = 0, limit: int = 100, search: str = None, title: str = None, variable: str = None):

    # sanitize the search
    if search is not None and search.strip() == '':
        search = None

    # call the function
    entries = core.entries(offset, limit, search=search,  title=title, variable=variable) 
    
    # check if we should return html
    if fmt == 'html':
        return templates.TemplateResponse(request=request, name="entries.html", context={"entries": entries})
    return entries


@app.get('/entries/new.html')
def new_entry_page(request: Request):
    return templates.TemplateResponse(request=request, name="add_entry.html", context={})


@app.post('/entries')
def add_entry(payload: models.EntryCreate) -> models.Metadata:
    metadata = core.add_entry(payload)
    return metadata


@app.get('/entries/{id}')
@app.get('/entries/{id}.{fmt}')
def get_entry(id: int, request: Request, fmt: Literal['xml'] | FMT = None):
    # call the function
    entries = core.entries(ids=id)
    
    if len(entries) == 0:
        return HTTPException(status_code=404, detail=f"Entry of <ID={id}> not found")
    
    # check if we should return html
    if fmt == 'html':
        return templates.TemplateResponse(request=request, name="entry.html", context={"entry": entries[0]})
    elif fmt == 'xml':
        return templates.TemplateResponse(request=request, name="entry.xml", context={"entry": entries[0]}, media_type='application/xml')
    return entries[0]


@app.get('/locations.{fmt}', response_model=FeatureCollectionModel)
def get_entries_geojson(request: Request, fmt: FMT = None, search: str = None, offset: int = None, limit: int = None, ids: int | List[int] = None):
    # check if we should return html
    if fmt == 'html':
        return templates.TemplateResponse(request=request, name="map.html", context={})
    
    # in all other casese call the function and return the feature collection
    geometries = core.entries_locations(limit=limit, offset=offset, search=search, ids=ids)
    
    return geometries


@app.get('/licenses')
@app.get('/licenses.{fmt}')
def get_licenses(request: Request, fmt: FMT = None, license_id: Optional[int] = None, variant: Literal['list', 'select'] = 'select'):
    # call the function
    try:
        licenses = core.licenses(id=license_id)
    except Exception as e:
         return HTTPException(status_code=404, detail=str(e))

    # check if we should return html
    if fmt == 'html':
        # check the number if a id was given
        if license_id is not None:
            return templates.TemplateResponse(request=request, name="license.html", context={"license": licenses.model_dump()})
        else:
            return templates.TemplateResponse(request=request, name="licenses.html", context={"licenses": licenses, "variant": variant})
    else:
        return licenses


@app.get('/authors/new.html')
def new_author(request: Request):
    return templates.TemplateResponse(request=request, name="author.html", context={})


@app.get('/authors')
@app.get('/authors.{fmt}')
@app.get('/entries/{entry_id}/authors.{fmt}')
def get_authors(request: Request, fmt: FMT = None, entry_id: int = None, author_id: int = None, search: str = None, exclude_ids: List[int] = None, target: str = None, offset: int = None, limit: int = None):
    try:
        authors = core.authors(id=author_id, entry_id=entry_id, search=search, exclude_ids=exclude_ids, offset=offset, limit=limit)
    except Exception as e:
        return HTTPException(status_code=404, detail=str(e))

    if fmt == 'html':
        if author_id is not None:
            return templates.TemplateResponse(
                request=request,
                name="author.html",
                context={"author": authors, 'variant': 'fixed', 'target': target}
            )
        return templates.TemplateResponse(
            request=request, 
            name="authors.html", 
            context={"authors": authors, 'variant': 'select' if entry_id is None else 'list', 'target': target}
        )
    else:
        return authors


@app.get('/authors/{author_id}')
@app.get('/authors/{author_id}.json')
def get_author(author_id: int, request: Request):
    try:
        author = core.authors(id=author_id)
        return author
    except Exception as e:
        return HTTPException(status_code=404, detail=str(e))


@app.get('/variables')
@app.get('/variables.{fmt}')
def get_variables(request: Request, fmt: FMT = None, offset: int = None, limit: int = None):
    if fmt == 'html':
        return templates.TemplateResponse(request=request, name="variables.html", context={})
    try:
        variables = core.variables(only_available=False, offset=offset, limit=limit)
    except Exception as e:
        return HTTPException(status_code=404, detail=str(e))
    
    if fmt == 'html':
        return templates.TemplateResponse(
            request=request, 
            name="variables.html", 
            context={"variables": variables}
        )
    else:
        return variables


@app.get('/details/new.html')
def new_details(request: Request):
    return templates.TemplateResponse(request=request, name="details.html", context={})


@app.get('/datasources/new.html')
def new_datasource(request: Request):
    # load the datasource types
    types = core.datatypes()
    return templates.TemplateResponse(request=request, name="add_datasource.html", context={"types": types})


@app.post('/entries/{entry_id}/datasource')
def add_datasource(entry_id: int, payload: models.DatasourceCreate) -> models.Metadata:
    metadata = core.add_datasource(entry_id=entry_id, payload=payload)
    return metadata

# add static files
app.mount("/static", StaticFiles(directory=Path(__file__).parent / "templates" / "static"), name="static")

# register the editing tools
app.include_router(editing_tools.edit_router, prefix='/utils')

if __name__ == "__main__":
    import uvicorn
    uvicorn.run('server:app', host="0.0.0.0", port=8000, reload=True)



from typing import List
from pathlib import Path

from sqlmodel import Session, text
from psycopg2.errors import UndefinedTable
from pydantic_geojson import FeatureCollectionModel
from pydantic import BaseModel

from metacatalog_api import models
from sqlmodel import select, exists, col

DB_VERSION = 2
SQL_DIR = Path(__file__).parent / "sql"

# helper function to load sql files
def load_sql(file_name: str) -> str:
    path = Path(file_name)
    if not path.exists():
        path = SQL_DIR / file_name
    
    with open(path, 'r') as f:
        return f.read()


# helper function to check the database version
def get_db_version(session: Session) -> dict:
    try:
        v = session.exec(text("SELECT db_version FROM metacatalog_info order by db_version desc limit 1;")).scalar() 
    except UndefinedTable:
        v = 0
    return {'db_version': v}


def check_db_version(session: Session) -> bool:
    """Verify that the database version matches the expected version.
    
    Args:
        session: A SQLAlchemy session object
        
    Raises:
        ValueError: If database version doesn't match DB_VERSION constant

    Returns:
        bool: True if database version matches
    """
    remote_db_version = get_db_version(session)['db_version']
    if remote_db_version != DB_VERSION:
        raise ValueError(
            f"Database version mismatch. Expected {DB_VERSION}, got {remote_db_version}. "
            "Please run database migrations to update your schema."
        )
    return True


def install(session: Session, schema: str = 'public', populate_defaults: bool = True) -> None:
    # get the install script
    install_sql = load_sql(SQL_DIR / 'maintain' /'install.sql').format(schema=schema)

    # execute the install script
    session.exec(text(install_sql))

    # populate the defaults
    if populate_defaults:
        pupulate_sql = load_sql(SQL_DIR / 'maintain' / 'defaults.sql').replace('{schema}', schema)
        session.exec(text(pupulate_sql))


def check_installed(session: Session, schema: str = 'public') -> bool:
    try:
        info = session.exec(text(f"SELECT * FROM information_schema.tables WHERE table_schema = '{schema}' AND table_name = 'entries'")).first() 
        return info is not None
    except Exception:
        return False
    

def get_entries(session: Session, limit: int = None, offset: int = None, variable: str | int = None, title: str = None) -> list[models.Metadata]:
    # build the base query
    sql = select(models.EntryTable)

    # handle variable filter
    if isinstance(variable, int):
        sql = sql.join(models.VariableTable).where(models.VariableTable.id == variable)
    elif isinstance(variable, str):
        sql = sql.join(models.VariableTable).where(col(models.VariableTable.name).ilike(variable))
    
    # handle title filter
    if title is not None:
        sql = sql.where(col(models.EntryTable.title).ilike(title))
    
    # handle offset and limit
    sql = sql.offset(offset).limit(limit)

    # execute the query
    entries = session.exec(sql).all()  

    return [models.Metadata.model_validate(entry) for entry in entries]


def get_entries_by_id(session: Session, entry_ids: int | list[int], limit: int = None, offset: int = None) -> list[models.Metadata] | models.Metadata:
    # base query
    sql = select(models.EntryTable)

    # handle entry ids
    if isinstance(entry_ids, int):
        sql = sql.where(models.EntryTable.id == entry_ids)
    elif isinstance(entry_ids, (list, tuple)):
        sql = sql.where(col(models.EntryTable.id).in_(entry_ids))

    # handle offset and limit
    sql = sql.offset(offset).limit(limit)

    # run the query
    entries = session.exec(sql).all()

    if isinstance(entries, models.EntryTable):
        return models.Metadata.model_validate(entries)
    else:
        return [models.Metadata.model_validate(entry) for entry in entries]


def get_entries_locations(session: Session, ids: List[int] = None, limit: int = None, offset: int = None) -> FeatureCollectionModel:
    # build the id filter
    if ids is None or len(ids) == 0:
        filt = ""
    else:
        filt = f" AND entries.id IN ({', '.join([str(i) for i in ids])})"
    
    # build limit and offset
    lim = f" LIMIT {limit} " if limit is not None else ""
    off = f" OFFSET {offset} " if offset is not None else ""

    # load the query
    sql = load_sql("entries_locations.sql").format(filter=filt, limit=lim, offset=off)

    # execute the query
    result = session.exec(text(sql)).one()[0]
        
    if result['features'] is None:
        return dict(type="FeatureCollection", features=[])
    
    return result
    

class SearchResult(BaseModel):
    id: int
    matches: list[str]
    weight: int


def search_entries(session: Session, search: str, limit: int = None, offset: int = None) -> list[SearchResult]:
    # build the limit and offset
    lim = f" LIMIT {limit} " if limit is not None else ""
    off = f" OFFSET {offset} " if offset is not None else ""

    # get the sql for the query
    sql = load_sql("search_entries.sql").format(prompt=search, limit=lim, offset=off)

    # execute the query
    mappings = session.exec(text(sql)).mappings().all()
    result = [m['search_meta'] for m in mappings]
    search_results = [SearchResult.model_validate(res) for res in result]

    return search_results


def get_authors(session: Session, search: str = None, exclude_ids: list[int] = None, limit: int = None, offset: int = None) -> List[models.Author]:
    # build the base query
    query = select(models.PersonTable)

    # handle search
    if search is not None:
        search = search.replace('*', '%')
        
        query = query.where(
            col(models.PersonTable.first_name).contains(search) |
            col(models.PersonTable.last_name).contains(search) |
            col(models.PersonTable.organisation_name).contains(search)
        )
    
    # handle exclude
    if exclude_ids is not None:
        query = query.where(
            col(models.PersonTable.id).not_in(exclude_ids)
        )
    
    # hanlde limit and offset
    query = query.offset(offset).limit(limit)

    # run
    authors = session.exec(query).all()

    return [models.Author.model_validate(author) for author in authors]


def get_authors_by_entry(session: Session, entry_id: int) -> list[models.Author]:
    query = (
        select(models.PersonTable).where(models.PersonTable.id.in_(
                select(models.EntryTable.author_id)
                .where(models.EntryTable.id == entry_id)
            )
        ).union_all(
            select(models.PersonTable).join(models.NMPersonEntries)
            .where(models.NMPersonEntries.entry_id == entry_id)
            .order_by(col(models.NMPersonEntries.order).asc()) 
        )
    )
    authors = session.exec(query).all()

    return [models.Author.model_validate(author) for author in authors]


def get_author_by_id(session: Session, id: int) -> models.Author:   
    # execute the query
    author = session.exec(select(models.PersonTable).where(models.PersonTable.id == id)).first()

    if author is None:
        raise ValueError(f"Author with id {id} not found")
    else:
        return models.Author.model_validate(author)


def get_variables(session: Session, limit: int = None, offset: int = None) -> list[models.Variable]:
    variables = session.exec(
        select(models.VariableTable).offset(offset).limit(limit)
    )

    return [models.Variable.model_validate(var) for var in variables]

    
def get_available_variables(session: Session, limit: int = None, offset: int = None) -> list[models.Variable]:
    # build the query
    query = select(models.VariableTable).where(
        exists(select(models.EntryTable.id).where(models.EntryTable.variable_id == models.VariableTable.id))
    ).offset(offset).limit(limit)
    
    # execute the query
    variables = session.exec(query).all()

    return [models.Variable.model_validate(var) for var in variables]


def get_variable_by_id(session: Session, id: int) -> models.Variable:
    variable = session.get(models.VariableTable, id)

    if variable is None:
        raise ValueError(f"Variable with id {id} not found")
    
    return models.Variable.model_validate(variable)


def get_licenses(session: Session, limit: int = None, offset: int = None) -> List[models.License]:
    # get the licenses
    licenses = session.exec(
        select(models.LicenseTable).offset(offset).limit(limit)
    ).all()

    return [models.License.model_validate(lic) for lic in licenses]


def get_license_by_id(session: Session, id: int) -> models.License:
    # get the one license in question
    lic = session.get(models.LicenseTable, id)
    if lic is None:
        raise ValueError(f"License with id {id} not found")
    else:
        return models.License.model_validate(lic)


def get_datatypes(session: Session, id: int = None) -> list[models.DatasourceTypeBase]:
    # handle the id
    if id is not None:
        sql = select(models.DatasourceTypeTable).where(models.DatasourceTypeTable.id == id)
        type_ = session.exec(sql).one()
        return models.DatasourceTypeBase.model_validate(type_)
    else:
        # get all the types
        types = session.exec(select(models.DatasourceTypeTable)).all()

        return [models.DatasourceTypeBase.model_validate(type_) for type_ in types]


# def get_datasource_by_id(session: Cursor, id: int) -> models.Datasource:
#     # handle the filter
#     raise NotImplementedError


def add_entry(session: Session, payload: models.EntryCreate) -> models.Metadata:
    # grab the keywords
    if payload.keywords is not None and len(payload.keywords) > 0:
        sql = select(models.KeywordTable).where(col(models.KeywordTable.id).in_(payload.keywords))
        keywords =  session.exec(sql).all()
    else:
        keywords = []
    
    # add or set the author
    if isinstance(payload.author, int):
        author = session.get(models.PersonTable, payload.author)
    else:
        author = models.PersonTable.model_validate(payload.author)
    
    # handle co-authors
    if payload.coAuthors is None or len(payload.coAuthors) == 0:
        coAuthors = []
    else:
        coAuthors = []
        for coAuthor in payload.coAuthors:
            if isinstance(coAuthor, int):
                coAuthors.append(session.get(models.PersonTable, coAuthor))

            else:
                coAuthors.append(models.PersonTable.model_validate(coAuthor))
    
    # handle license
    if isinstance(payload.license, int):
        license = session.get(models.LicenseTable, payload.license)
    else:
        license = models.LicenseCreate.model_validate(payload.license)

    # create the table entry
    entry = models.EntryTable(
        title=payload.title,
        abstract=payload.abstract,
        external_id=payload.external_id,
        #location=payload.location,
        version=payload.version,
        is_partial=payload.is_partial,
        comment=payload.comment,
        citation=payload.citation,
        #embargo=payload.embargo,
        #embargo_end=payload.embargo_end,
        license=license,
        variable_id=payload.variable,
        author=author,
        coAuthors=coAuthors,
        keywords=keywords
    )
    if payload.location is not None:
        entry.location = models.EntryTable.validate_location(payload.location, None)

    # add
    session.add(entry)
    session.commit()

        # build the details
    if payload.details is not None and len(payload.details) > 0:
        details = []
        for d in payload.details:
            details.append(models.DetailTable(
                key=d.key, 
                raw_value=d.raw_value,
                entry_id=entry.id,
                thesaurus_id=d.thesaurus,
                title=d.title,
                description=d.description
            ))
        entry.details = details
        session.add(entry)
        session.commit()

    # refresh the entry object and validate the Metadata model
    session.refresh(entry)
    return models.Metadata.model_validate(entry)


def add_datasource(session: Session, entry_id: int, datasource: models.DatasourceCreate) -> models.Metadata:
    # get the entry
    entry = session.get(models.EntryTable, entry_id)
    if entry is None:
        raise ValueError(f"Entry with id {entry_id} not found")
    # look up the datasource type
    if isinstance(datasource.type, str):
        sql = select(models.DatasourceTypeTable.id).where(col(models.DatasourceTypeTable.name) == datasource.type)
    else:
        sql = select(models.DatasourceTypeTable.id).where(models.DatasourceTypeTable.id == datasource.type)
    
    # get the datasource type id
    datasource_type_id = session.exec(sql).first()
    if datasource_type_id is None:
        raise ValueError(f"Datasource type with name or id {datasource.type} was not found in the database")
    
    # check if a temporal scale is provided
    if datasource.temporal_scale is not None:
        temporal_scale = models.TemporalScaleTable.model_validate(datasource.temporal_scale)
    else:
        temporal_scale = None
    
    # check if a spatial scale is provided
    if datasource.spatial_scale is not None:
        spatial_scale = models.SpatialScaleTable.model_validate(datasource.spatial_scale)
    else:
        spatial_scale = None

    # create the table entry
    datasource = models.DatasourceTable(
        path=datasource.path,
        encoding=datasource.encoding,
        type_id=datasource_type_id,
        args=datasource.args if datasource.args is not None else {},
        temporal_scale=temporal_scale,
        spatial_scale=spatial_scale,
        variable_names=datasource.variable_names if datasource.variable_names is not None else []
    )

    # add the datasource
    entry.datasource = datasource
    session.add(entry)
    session.commit()

    session.refresh(entry)
    return models.Metadata.model_validate(entry)



from typing import List, Generator
import os
from pathlib import Path
from contextlib import contextmanager

from sqlmodel import Session, create_engine, text
from metacatalog_api import models
from dotenv import load_dotenv
from pydantic_geojson import FeatureCollectionModel

from metacatalog_api import db
#from metacatalog_api import __version__ as metacatalog_version


load_dotenv()

METACATALOG_URI = os.environ.get("METACATALOG_URI", 'postgresql://metacatalog:metacatalog@localhost:5432/metacatalog')
SQL_DIR = Path(__file__).parent / "sql"


@contextmanager
def connect(url: str = None) -> Generator[Session, None, None]:
    uri = url if url is not None else METACATALOG_URI
    engine = create_engine(uri)

    with Session(engine) as session:
        yield session

def get_session(url: str = None) -> Session:
    if url is None:
        url = os.getenv('METACATALOG_URI')
        
    engine = create_engine(url)
    return Session(engine)


def migrate_db(schema: str = 'public') -> None:
    # get the current version
    with connect() as session:
        current_version = db.get_db_version(session)['db_version']
    
    # as long as the local _DB_VERSION is higher than the remote version, we can load a migration
    if current_version < db.DB_VERSION:
        # get the migration sql
        migration_sql = db.load_sql(SQL_DIR / 'migrate' / f'migration_{current_version + 1}.sql').format(schema=schema)
        
        # run
        with connect() as session:
            session.exec(text(migration_sql))
            
            # update the db version
            session.exec(text(f"INSERT INTO metacatalog_info (db_version) VALUES ({current_version + 1});")) 
            session.commit()
        # inform the user
        print(f"Migrated database from version {current_version} to {current_version + 1}")
        
        # finally call the migration function recursively
        migrate_db()
        

def entries(offset: int = 0, limit: int = 100, ids: int | List[int] = None,  search: str = None, variable: str | int = None, title: str = None) -> list[models.Metadata]:
    # check if we filter or search
    with connect() as session:
        if search is not None:
            search_results = db.search_entries(session, search, limit=limit, offset=offset)

            if len(search_results) == 0:
                return []
            # in any other case get them by id
            # in any other case, request the entries by id
            results = db.get_entries_by_id(session=session, entry_ids=[r.id for r in search_results])

            return results
        elif ids is not None:
            results = db.get_entries_by_id(session, ids, limit=limit, offset=offset)
        else:
            results = db.get_entries(session, limit=limit, offset=offset, variable=variable, title=title)

    return results


def entries_locations(ids: int | List[int] = None, limit: int = None, offset: int = None, search: str = None, filter: dict = {}) -> FeatureCollectionModel:
    # handle the ids
    if ids is None:
        ids = []
    if isinstance(ids, int):
        ids = [ids]
    
    # check if we filter or search
    with connect() as session:
        # run the search to ge the ids
        if search is not None:
            search_results = db.search_entries(session, search, limit=limit, offset=offset)
            ids = [*ids, *[r.id for r in search_results]]
        
            # if no ids have been found, return an empty FeatureCollection
            if len(ids) == 0:
                return {"type": "FeatureCollection", "features": []}
        
        # in any other case we go for the locations.
        result = db.get_entries_locations(session, ids=ids, limit=limit, offset=offset)
    
    return result


def licenses(id: int = None, offset: int = None, limit: int = None) -> models.License | list[models.License]:
    with connect() as session:
        if id is not None:
            result = db.get_license_by_id(session, id=id)
        else:
            result = db.get_licenses(session, limit=limit, offset=offset)
    
    return result


def authors(id: int = None, entry_id: int = None, search: str = None, exclude_ids: List[int] = None, offset: int = None, limit: int = None) -> List[models.Author]:
    with connect() as session:
        # if an author_id is given, we return only the author of that id
        if id is not None:
            authors = db.get_author_by_id(session, id=id)
        # if an entry_id is given, we return only the authors of that entry
        elif entry_id is not None:
            authors = db.get_authors_by_entry(session, entry_id=entry_id)
        else:
            authors = db.get_authors(session, search=search, exclude_ids=exclude_ids, limit=limit, offset=offset)
    
    return authors


def variables(id: int = None, only_available: bool = False, offset: int = None, limit: int = None) -> List[models.Variable]:
    with connect() as session:
        if only_available:
            variables = db.get_available_variables(session, limit=limit, offset=offset)
        elif id is not None:
            variables = [db.get_variable_by_id(session, id=id)]
        else:
            variables = db.get_variables(session, limit=limit, offset=offset)
    
    return variables


def datatypes(id: int = None) -> List[models.DatasourceTypeBase]:
    # TODO: this may need some more parameters
    with connect() as session:
        return db.get_datatypes(session, id=id)


def add_entry(payload: models.EntryCreate) -> models.Metadata:
    # add the entry
    with connect() as session:
        entry = db.add_entry(session, payload=payload)
    
        # check if there was a datasource
        if payload.datasource is not None:
            entry = db.add_datasource(session, entry_id=entry.id, datasource=payload.datasource)
        session.commit()
    return entry

def add_datasource(entry_id: int, payload: models.DatasourceCreate) -> models.Metadata:
    with connect() as session:
        entry = db.add_datasource(session, entry_id=entry_id, datasource=payload)

    return entry